---------------------------------------------------------------------------
-- HIERARCHIAL CONFIG MANAGEMENT CLASSS
local HIER_CONFIG = {};

function HIER_CONFIG:GetEffectiveConfig(groupId, defaultCopy, resultTable)
   local config = self[groupId];
   if ((not config) and (defaultCopy)) then
      config = self[defaultCopy];
   end
   -- Sanity check in case table is loopily self-referential
   local escapeCount = 100;
   while (config) do
      local copyOf = config.copyOf;
      if (not copyOf) then 
	 break;
      end
      escapeCount = escapeCount - 1;
      if (escapeCount < 0) then
	 return nil;
      end
      config = self[copyOf] or (defaultCopy and self[defaultCopy]);
   end

   if ((not resultTable) or (not config)) then return config; end
   for k in pairs(resultTable) do
      resultTable[k] = nil;
   end
   
   for k,v in pairs(config) do
      resultTable[k] = v;
   end

   return resultTable;
end

function HIER_CONFIG:GetConfig(groupId, defaultId, resultTable)
   local ret = self[groupId];
   if (not ret) then
      if ((not resultTable) or (not defaultId)) then 
	 return nil;
      end
      for k in resultTable do
	 resultTable[k] = nil;
      end
      if (defaultId) then
	 resultTable.copyOf = defaultId;
      end
      return resultTable;
   end
   if (not resultTable) then return ret; end
   for k in resultTable do
      resultTable[k] = nil;
   end
   for k,v in ret do
      resultTable[k] = v;
   end
   return resultTable;
end

local function HS_CCE_Recurse(configTable, groupId, workMap, defaultId)
   local cur = workMap[groupId];
   if (cur ~= nil) then return cur; end

   -- Unknown at this point
   local config = configTable[groupId];
   local copyOf;

   if (not config) then
      if (not defaultId) then
	 -- Dead end!
	 workMap[groupId] = false;
	 return false;
      else
	 -- Pretend this was a pass through to defaultId
	 copyOf = defaultId;
      end
   else
      copyOf = config.copyOf;
      if (not copyOf) then
	 -- Excellent, this is where we get off!
	 workMap[groupId] = true;
	 return true;
      end
   end
   
   -- Prevent looping
   workMap[groupId] = false;
   
   -- Descend
   cur = HS_CCE_Recurse(configTable, copyOf, workMap, defaultId);
   if (cur) then workMap[groupId] = true; end
   return cur;
end

local HS_CCE_WorkMap = {};

function HIER_CONFIG:CheckCopyEligibility(groupId, groupMap, defaultId)
   for k in pairs(HS_CCE_WorkMap) do 
      HS_CCE_WorkMap[k] = nil;
   end
   HS_CCE_WorkMap[groupId] = false;
   for k in pairs(groupMap) do
      local cur = HS_CCE_WorkMap[k];
      if (cur == nil) then
	 cur = HS_CCE_Recurse(self, k, HS_CCE_WorkMap, defaultId);
      end
      groupMap[k] = cur;
   end
   return groupMap;
end

local function HS_FDG_Recurse(configTable, groupId, workMap, defaultId)
   local cur = workMap[groupId];
   if (cur ~= nil) then return cur; end

   -- Unknown at this point
   local config = configTable[groupId];
   local copyOf;

   if (not config) then
      if (not defaultId) then
	 -- Dead end!
	 workMap[groupId] = false;
	 return false;
      else
	 -- Pretend this was a pass through to defaultId
	 copyOf = defaultId;
      end
   else
      copyOf = config.copyOf;
      if (not copyOf) then
	 -- Solid end, not where we needed to be
	 workMap[groupId] = false;
	 return false;
      end
   end
   
   -- Prevent looping
   workMap[groupId] = false;
   
   -- Descend
   cur = HS_FDG_Recurse(configTable, copyOf, workMap, defaultId);
   if (cur) then workMap[groupId] = true; end
   return cur;
end

function HIER_CONFIG:FindDependentGroups(groupId, groupMap, defaultId)
   for k in pairs(HS_CCE_WorkMap) do 
      HS_CCE_WorkMap[k] = nil;
   end
   HS_CCE_WorkMap[groupId] = true;
   for k in pairs(groupMap) do
      local cur = HS_CCE_WorkMap[k];
      if (cur == nil) then
	 cur = HS_FDG_Recurse(self, k, HS_CCE_WorkMap, defaultId);
      end
      groupMap[k] = cur;
   end
   return groupMap;
end

function HIER_CONFIG:ApplyConfig(groupId, config, defaultId, 
				     noDefaultNil)
   local copyOf = config.copyOf;
   if (copyOf) then
      -- Cycle detection
      local copyTmp = copyOf;
      for k in pairs(HS_CCE_WorkMap) do 
	 HS_CCE_WorkMap[k] = nil;
      end
      HS_CCE_WorkMap[groupId] = true;
      while (copyTmp) do
	 if (HS_CCE_WorkMap[copyTmp]) then
	    -- Cycle detected, bye bye
	    return nil;
	 end
	 HS_CCE_WorkMap[copyTmp] = true;
	 local subSet = self[copyOf];
	 if ((not subSet) and (defaultId)) then
	    subSet = self[defaultId];
	 end
	 if (not subSet) then
	    -- References non-existant setting
	    return nil;
	 end
	 copyTmp = subSet.copyOf;
      end

      -- If we're setting to copy default ID, then just nil out the
      -- setting
      if ((copyOf == defaultId) and (not noDefaultNil)) then
	 self[groupId] = nil;
	 return true;
      end
   end

   local curSet = self[groupId];
   if (curSet) then
      for k in pairs(curSet) do 
	 curSet[k] = nil;
      end
   else
      curSet = {};
      self[groupId] = curSet;
   end
   for k,v in pairs(config) do
      curSet[k] = v;
   end
   return true;
end

function StatRingsConfig_BlessHierarchialConfig(object)
   local metatbl = {};
   metatbl.__index = HIER_CONFIG;
   setmetatable(object, metatbl);
   return object;
end

