-- LUA code for scaffold configuration GUI

local function DEBUG(msg) 
   DEFAULT_CHAT_FRAME:AddMessage("[SROS] " .. msg);
end

local SRO_ROW = {};

function SRO_ROW:SROAction(subframe, detail)
   local id = self.id;
   if (not id) then return; end
   local op;
   if (subframe == self.editButton) then
      op = "edit";
   elseif (subframe == self.moveButton) then
      op = "move";
   elseif (subframe == self.checkButton) then
      op = "visible";
   else
      return;
   end
   local det="";
   if (detail ~= nil) then
      det = (detail and " true") or " false";
   end
   DEFAULT_CHAT_FRAME:AddMessage("[" .. id .. "] "
				 .. (op or "none") .. det);

   if (op == "visible") then
      if (self.shownState ~= nil) then
	 self.shownState = (detail and true) or false;
	 if (detail) then
	    self.moveButton:Enable();
	 else
	    self.moveButton:Disable();
	 end
      end
   elseif (op == "move") then
      self.ownerFrame:EditAnchor(id);
   end
end


function SRO_ROW:SetInfo(id, label, rowNum, shownState, canEdit, moveLabel)
   label = label or id;
   self.id = id;
   self.label = label;
   if (not label) then
      self:Hide();
      self.rowNum = nil;
      return;
   else
      self.labelText:SetText(label);
      self:Show();
   end
   self.rowNum = rowNum;
   if (math.mod(rowNum,2) == 0) then
      self.bgTexture:Show();
   else
      self.bgTexture:Hide();
   end
   self.shownState = shownState;
   if (shownState == nil) then
      self.checkButton:Hide();
      self.checkButton:SetChecked(true);
   else
      self.checkButton:Show();
      self.checkButton:SetChecked(shownState);
   end
   if (canEdit) then
      self.editButton:Show();
   else
      self.editButton:Hide();
   end
   if (moveLabel) then
      if ((shownState ~= nil) and not shownState) then
	 self.moveButton:Disable();
      else
	 self.moveButton:Enable();
      end
	    
      self.moveButton:SetText(moveLabel);
      self.moveButton:Show();
   else
      self.moveButton:Hide();
   end
end

function StatRingsOptionsScaffold_Row_OnLoad(this)
   local name = this:GetName();
   local editButton = getglobal(name .. "BE");
   local moveButton = getglobal(name .. "BM");
   local checkButton = getglobal(name .. "CB");
   local labelText = getglobal(name .. "F");
   local bgTexture = getglobal(name .. "T");

   this.name = name;
   this.editButton = editButton;
   this.moveButton = moveButton;
   this.checkButton = checkButton;
   this.labelText = labelText;
   this.bgTexture = bgTexture;

   for methodName, func in pairs(SRO_ROW) do
      this[methodName] = func;
   end
end

---------------------------------------------------------------------------
-- The options frame itself
local SRO_FRAME = {};

function SRO_FRAME:ScrollUpdate()
   local scrollFrame = self.scrollFrame;
   local offset = FauxScrollFrame_GetOffset(scrollFrame);

   local dataCount = table.getn(self.ids);
   local rowCount = self.rowCount;
   local rowHeight = self.rowHeight;

   FauxScrollFrame_Update(scrollFrame, dataCount, rowCount, rowHeight);
   local frameWidth = 326;
   if (not scrollFrame:IsShown()) then
      frameWidth = frameWidth + 26;
   end

   if (offset ~= self.curOffset) then
      local sfName = scrollFrame:GetName();
      self.curOffset = offset;
      local minShown = offset + 1;
      local maxShown = minShown + rowCount - 1;
      if (maxShown > dataCount) then
	 maxShown = dataCount;
      end
      local work = self.updateWork;
      local rows = self.rowFrames;

      local freePtr = 0;

      for i=1,rowCount do
	 -- First scan the visible rows and relocate those we still want
	 local row = rows[i];
	 rows[i] = false; -- Clear up for next time
	 local rowNum = row.rowNum or -1;
	 
	 if ((rowNum < minShown) or (rowNum > maxShown)) then
	    freePtr = freePtr + 1;
	    rows[freePtr] = row;
	    -- Mark this one as missing
	 else
	    -- Store row offset from top
	    work[rowNum - minShown + 1] = row;
	 end
      end
      freePtr = 0;
      for i=1,rowCount do
	 -- Next assign unused ones and anchor/hide
	 local row = work[i];
	 if (not row) then
	    freePtr = freePtr + 1;
	    row = rows[freePtr];
	    rows[freePtr] = false; -- Clear up for next time
	    work[i] = row;

	    local rowNum = minShown + i - 1;
	    local id;
	    if ((rowNum <= maxShown) and (rowNum >= minShown)) then
	       id = self.ids[rowNum];
	    end
	    if (id) then
	       self:ConfigureRow(row, rowNum, id);
	    else
	       row:SetInfo(nil);	   
	    end
	 end

	 if (row.rowNum) then
	    row:SetWidth(frameWidth);
	    row:SetPoint("TOPLEFT", sfName, "TOPLEFT", 0, 
			 (1-i) * rowHeight + 2);
	 end
      end

      self.updateWork, self.rowFrames = rows, work;
   end
end

function SRO_FRAME:ConfigureRow(row, rowNum, id) 
   if (not self.scaffold) then
      row:SetInfo(nil);
      return;
   end
   local owner = self.owners[id];
   local scaffold = self.scaffold;
   local nodeType,isBroken = scaffold:GetType(id);

   local label, shownState, canEdit, moveLabel = id, nil, false, "View";
   if (scaffold:CanSetAnchor(id)) then
      moveLabel = "Move";
   end
   
   if (nodeType == "root") then
      label = "Root: " .. id;
   elseif (nodeType == "group") then
      label = "Group: " .. id;
      canEdit = true;
      shownState = true;
      moveLabel = nil;
   else
      if (owner) then
	 if (nodeType == "member") then
	    label = "    Member: " .. id;
	    shownState = true;
	 elseif (id == owner .. "Start") then
	    label = "    Group Start";
	 elseif (id == owner .. "End") then
	    label = "    Group End";
	 else
	    -- This should never happen
	    label = "    Node: " .. id;
	 end
      else
	 label = "Node: " .. id;
      end
	    
      if (nodeType == "node") then
	 canEdit = true;
	 shownState = true;
      end
   end

   if (isBroken) then
      label = label .. " !";
      moveLabel = nil;
      shownState = nil;
   end
   
   row:SetInfo(id, label, rowNum, shownState, canEdit, moveLabel);
end

function SRO_FRAME:SetScaffold(scaffold)
   self.scaffold = scaffold;
   local rows = self.rowFrames;
   for i=1, self.rowCount do
      if (rows[i].rowNumber) then
	 rows[i]:SetInfo(nil);
      end
   end
   self.curOffset = nil;
   FauxScrollFrame_SetOffset(self.scrollFrame, 0);
   if (not scaffold) then
      self.ids = {};
      self:ScrollUpdate();
      return;
   end

   -- Work out the ids
   local allIds = {};
   local nodeTypes = {};
   local num = scaffold:GetNumNodes();
   for i=1,num do
      local id = scaffold:GetNodeName(i);
      local nodeType = scaffold:GetType(id);
      nodeTypes[id] = nodeType;
      allIds[i] = id;
   end

   table.sort(allIds, function(a,b) 
			 local aT = nodeTypes[a];
			 local bT = nodeTypes[b];
			 if (aT ~= bT) then
			    if (aT == "group") then return; end
			    if (bT == "group") then return true; end
			 end
			 return (a < b);
		      end);

   local tinsert = table.insert;
   local ids = {};
   -- First, put the roots at the top of the list
   for i=1,num do
      local id = allIds[i];
      if (id and nodeTypes[id] == "root") then
	 tinsert(ids, id);
	 allIds[i] = nil;
	 nodeTypes[id] = nil;
      end
   end

   -- next remove group anchors from the list
   for i=1,num do
      local id = allIds[i];
      if (id and nodeTypes[id] == "group") then
	 -- Do group endpoints
	 local sName = id .. "Start";
	 local eName = id .. "End";
	 if (nodeTypes[sName] == "node") then
	    nodeTypes[sName] = "groupEndpoint";
	 end
	 if (nodeTypes[eName] == "node") then
	    nodeTypes[eName] = "groupEndpoint";
	 end
      end
   end

   local owners = {};
   local allDone = false;
   while (not allDone) do
      local anyWaiting, anyProcessed;
      for i=1,num do
	 local id = allIds[i];
	 if (id) then
	    local nodeType = nodeTypes[id];
	    local depa,depb = scaffold:GetDependencies(id);
	    if (depb == depa) then depb = nil; end
	    if ((depa and nodeTypes[depa]) or
		(depb and nodeTypes[depb])) then
	       anyWaiting = true;
	    else
	       nodeTypes[id] = nil;
	       allIds[i] = nil;
	       anyProcessed = true;

	       if (nodeType == "group") then
		  tinsert(ids, id);
		  local subid = id.."Start";
		  if (scaffold:GetType(subid) == "node") then
		     tinsert(ids, subid);
		     owners[subid] = id;
		  end
		  subid = id .. "End";
		  if (scaffold:GetType(subid) == "node") then
		     tinsert(ids, subid);
		     owners[subid] = id;
		  end
		  for j=1,scaffold:GetNumGroupMembers(id) do
		     subid = scaffold:GetGroupMember(id, j);
		     if ((scaffold:GetParent(subid) == id)
			 and (scaffold:GetType(subid) == "member")) then
			tinsert(ids, subid);
			owners[subid] = id;
		     end
		  end
	       elseif ((nodeType ~= "groupEndpoint") 
		   and (nodeType ~= "member")) then
		  tinsert(ids, id);
	       end

	       if (anyWaiting) then
		  break;
	       end
	    end
	 end
      end
      allDone = not (anyProcessed and anyWaiting);
   end
   
   -- all that fun stuff
   
   self.ids = ids;
   self.owners = owners;
   self:ScrollUpdate();
end

local ENABLED_PREVIEW = {
   shown = true,
   healthPct = 1,
   manaPct = 1,
   manaMax = 100,
   alpha = 1,
   unitName = true, -- Special value to show unit ID
};

local DISABLED_PREVIEW = {
   shown = true,
   force = "grey",
   alpha = 0.6,
   isGrey = true, -- Force show in greyscale
};

function SRO_FRAME:EditAnchor(id)
   local scaffold = StatRingsCluster.scaffold;

   if (self.editingAnchor) then
      local oldAnchor = self.editingAnchor;
      StatRingsOptionsDragFrame.drag.mode = nil;
      StatRingsOptionsDragFrame.drag:Hide();
      StatRingsOptionsDragFrame:Hide();
      StatRingsCluster:RestoreInfoModel();
      self.editingAnchor = nil;
      if (oldAnchor == id) then
	 return;
      end
   end

   if (not id) then
      return;
   end

   self.editingAnchor = id;
   
   local anchor = scaffold:GetAnchor(id);
   if (not anchor) then
      DEBUG("Unable to find anchor '" .. id .. "'");
      return;
   end
   
   local parentName = scaffold:GetParent(id);
   local x,y,scale,spin = scaffold:GetLocation(parentName);
   if (not x) then
      DEBUG("Anchor parent is not visible");
      return;
   end

   StatRingsCluster:PreviewInfoModel(function() return true; end,
				     ENABLED_PREVIEW,
				     DISABLED_PREVIEW);

   StatRingsOptionsDrag_Start("Adjusting " .. id,
			      StatRingsCluster,
			      x, y, scale / 100, spin,
			      anchor,
			      function(newAnchor) 
				 scaffold:SetAnchor(id, newAnchor);
				 scaffold:Update();
			      end);
end

function StatRingsOptionsScaffold_Frame_OnLoad(this)
   local name = this:GetName();
   local sfName = name .. "ScrollFrame";
   this.scrollFrame = getglobal(sfName);
   
   this.scrollFrame.owner = this;
   this.scrollFrame.fauxUpdate = function()
				    this:ScrollUpdate();
				 end;

   -- Copy in frame utility methods
   for methodName, func in pairs(SRO_FRAME) do
      this[methodName] = func;
   end

   local ids = {
--      "Node: player", "Node: pet", "Node: target",
--      "Group: party",
--      "    Group Start", "    Group End", 
--      "    Member: party1", "    Member: party2", 
--      "    Member: party3", "    Member: party4",
--      "Node: partypet",
--      "Node: partypet1",
--      "Node: partypet2",
--      "Node: partypet3",
--      "Node: partypet4"
   };

   this.ids = ids;
   
   local rows = {};
   this.rowFrames = rows;

   local rowHeight = 23;
   this.rowHeight = 23;

   local n = 0;
   while (true) do
      n = n + 1;
      local rName = name .. "R" .. n;
      local rFrame = getglobal(rName);
      if (not rFrame) then
	 break;
      end
      rows[n] = rFrame;
      rFrame.ownerFrame = this;
      rFrame:ClearAllPoints();
      rFrame:SetPoint("TOPLEFT", sfName, "TOPLEFT", 0, 
		      (1-n) * rowHeight + 2);
   end

   local rowCount = table.getn(rows);
   this.rowCount = rowCount;

   local work = {};
   this.updateWork = work;
   for n=1,rowCount do
      work[n] = false;
   end
end

function StatRingsOptionsScaffold_Frame_OnShow(this)
   -- Update the scroll bar.
   this:ScrollUpdate();
end


SlashCmdList["STATRINGS_OPTIONSSCAFFOLD"] = 
   function()
      if (SROScaffoldFrame:IsShown()) then
	 SROScaffoldFrame:EditAnchor(nil)
	 SROScaffoldFrame:Hide();
	 return;
      end
      SROScaffoldFrame:SetScaffold(StatRingsCluster.scaffold);
      SROScaffoldFrame:Show();
   end
SLASH_STATRINGS_OPTIONSSCAFFOLD1 = "/sros";
SLASH_STATRINGS_OPTIONSSCAFFOLD2 = "/srdrag";
